from __future__ import annotations

from typing import Any

from scitbx.array_family import flex

def dataset_as_flex_double(dataset_id: Any, selection: tuple[slice]) -> flex.double: ...
def dataset_as_flex_float(dataset_id: Any, selection: tuple[slice]) -> flex.float: ...
def dataset_as_flex_int(dataset_id: Any, selection: tuple[slice]) -> flex.int: ...
